// zeroVM - VM crash course
// How to build a simple virtual machine
// Stefan Pietzonke 2019
//

#include "global.h"


#define EXE_NEXT(); ep = ep + eoffs; goto *jumpt[code[ep]];


/* empty opcode, doing nothing */
#define NOP 255;


// protos
S2 load_object (U1 *name);
void free_modules (void);
size_t strlen_safe (const char * str, int maxlen);

S8 data_size ALIGN;
S8 code_size ALIGN;

// see global.h user settings on top
S8 max_code_size ALIGN = MAX_CODE_SIZE;
S8 max_data_size ALIGN = MAX_DATA_SIZE;

S8 data_mem_size ALIGN;
S8 stack_size ALIGN = STACKSIZE;		// stack size added to data size when dumped to object file

// code
U1 *code = NULL;

// data
U1 *data = NULL;


S8 code_ind ALIGN;
S8 data_ind ALIGN;
S8 modules_ind ALIGN = -1;    // no module loaded = -1

S8 cpu_ind ALIGN = 0;

S8 max_cpu ALIGN = MAXCPUCORES;    // number of threads that can be runned

U1 silent_run = 0;

struct data_info data_info[MAXDATAINFO];
S8 data_info_ind ALIGN = -1;


void cleanup (void)
{
    // free_modules ();
	if (data) free (data);
    if (code) free (code);
	// if (threaddata) free (threaddata);
}

// new: function run: vm runtime
S2 run (void *arg)
{
	// execution bytecode pointer
	S8 ep = 16;				// execution pointer
	S8 eoffs = 0;		// number of bytes to skip to get next opcode

	S8 regi[256];			// integer registers 64 bit
	S8 arg1, arg2, arg3;	// opcode arguments

	S8 cpu_core = 0;

	U1 *bptr;

	printf ("zeroVM\n");

	// jumptable for indirect threading execution
	// filling every still not defined part as &&nop, NO OPERATION
	static void *jumpt[] =
	{
		&&pushb, &&nop, &&nop, &&nop, &&nop,
		&&pullb, &&nop, &&nop, &&nop, &&nop,
		&&addi, &&subi, &&muli, &&divi,
		&&nop, &&nop, &&nop, &&nop,
		&&nop, &&nop,
		&&nop, &&nop, &&nop, &&nop, &&nop, &&nop,
		&&nop, &&nop, &&nop, &&nop, &&nop, &&nop,
		&&nop, &&nop, &&nop, &&nop, &&nop, &&nop,
		&&nop, &&nop,
		&&nop, &&nop, &&nop, &&nop, &&nop, &&nop,
		&&loada, &&nop,
		&&intr0, &&nop, &&nop, &&nop,
		&&nop, &&nop, &&nop, &&nop,
		&&nop, &&nop, &&nop, &&nop,
        &&nop
	};


	// jumptable for indirect threading execution

	/*
	static void *jumpt[] =
    {
		&&pushb, &&pullb, &&loadb, &&addi, &&subi, &&muli, &&divi, &&intr0
	};
	*/

	printf ("running test program...\n\n");

	// fetch first bytecode
	EXE_NEXT();

	nop:
	// doing nothing
	eoffs = 1;
	EXE_NEXT();


	pushb:
	#if DEBUG
	printf ("%lli PUSHB\n", cpu_core);
	#endif
	arg1 = regi[code[ep + 1]];
	arg2 = regi[code[ep + 2]];
	arg3 = code[ep + 3];

	regi[arg3] = data[arg1 + arg2];

	eoffs = 4;
	EXE_NEXT();


	pullb:
	#if DEBUG
	printf ("%lli PULLB\n", cpu_core);
	#endif
	arg1 = code[ep + 1];
	arg2 = regi[code[ep + 2]];
	arg3 = regi[code[ep + 3]];

	data[arg2 + arg3] = regi[arg1];

	eoffs = 4;
	EXE_NEXT();


	loada:
	#if DEBUG
	printf ("%lli LOADA\n", cpu_core);
	#endif
	// data
	bptr = (U1 *) &arg1;

	*bptr = code[ep + 1];
	bptr++;
	*bptr = code[ep + 2];
	bptr++;
	*bptr = code[ep + 3];
	bptr++;
	*bptr = code[ep + 4];
	bptr++;
	*bptr = code[ep + 5];
	bptr++;
	*bptr = code[ep + 6];
	bptr++;
	*bptr = code[ep + 7];
	bptr++;
	*bptr = code[ep + 8];

	// offset

	//printf ("arg1: %li\n", arg1);

	bptr = (U1 *) &arg2;

	*bptr = code[ep + 9];
	bptr++;
	*bptr = code[ep + 10];
	bptr++;
	*bptr = code[ep + 11];
	bptr++;
	*bptr = code[ep + 12];
	bptr++;
	*bptr = code[ep + 13];
	bptr++;
	*bptr = code[ep + 14];
	bptr++;
	*bptr = code[ep + 15];
	bptr++;
	*bptr = code[ep + 16];

	//printf ("arg2: %li\n", arg2);

	arg3 = code[ep + 17];

	bptr = (U1 *) &regi[arg3];

	*bptr = data[arg1 + arg2];
	bptr++;
	*bptr = data[arg1 + arg2 + 1];
	bptr++;
	*bptr = data[arg1 + arg2 + 2];
	bptr++;
	*bptr = data[arg1 + arg2 + 3];
	bptr++;
	*bptr = data[arg1 + arg2 + 4];
	bptr++;
	*bptr = data[arg1 + arg2 + 5];
	bptr++;
	*bptr = data[arg1 + arg2 + 6];
	bptr++;
	*bptr = data[arg1 + arg2 + 7];

	//printf ("LOADA: %li\n", regi[arg3]);

	eoffs = 18;
	EXE_NEXT();


	addi:
	#if DEBUG
	printf ("%lli ADDI\n", cpu_core);
	#endif
	arg1 = code[ep + 1];
	arg2 = code[ep + 2];
	arg3 = code[ep + 3];

	regi[arg3] = regi[arg1] + regi[arg2];

	eoffs = 4;
	EXE_NEXT();


	subi:
	#if DEBUG
	printf ("%lli SUBI\n", cpu_core);
	#endif
	arg1 = code[ep + 1];
	arg2 = code[ep + 2];
	arg3 = code[ep + 3];

	regi[arg3] = regi[arg1] - regi[arg2];

	eoffs = 4;
	EXE_NEXT();


	muli:
	#if DEBUG
	printf ("%lli MULI\n", cpu_core);
	#endif
	arg1 = code[ep + 1];
	arg2 = code[ep + 2];
	arg3 = code[ep + 3];

	regi[arg3] = regi[arg1] * regi[arg2];

	eoffs = 4;
	EXE_NEXT();


	divi:
	#if DEBUG
	printf ("%lli DIVI\n", cpu_core);
	#endif
	arg1 = code[ep + 1];
	arg2 = code[ep + 2];
	arg3 = code[ep + 3];

	regi[arg3] = regi[arg1] / regi[arg2];

	eoffs = 4;
	EXE_NEXT();


	intr0:
	arg1 = code[ep + 1];
	#if DEBUG
	printf ("%lli INTR0: %lli\n", cpu_core, arg1);
	#endif
	switch (arg1)
	{
		case 4:
			//printf ("PRINTI\n");
			arg2 = code[ep + 2];
			printf ("%lli", regi[arg2]);
			eoffs = 5;
			break;

		case 7:
			//printf ("PRINTNEWLINE\n");
			printf ("\n");
			eoffs = 5;
			break;

		case 8:
			//	DELAY
			arg2 = code[ep + 2];
			usleep (regi[arg2] * 1000);
			eoffs = 5;
			break;

		case 255:
			printf ("shutdown...\n");
			return (0);
	}
	EXE_NEXT();
}

int main (int ac, char *av[])
{
	if (ac > 1)
	{
		if (load_object ((U1 *) av[1]))
    	{
			cleanup ();
        	exit (1);
    	}
	}
	else
	{
		printf ("zerovm-0.3 <program>\n");
		exit (1);
	}
	run (NULL);
	cleanup ();
	exit (0);
}
